/* =============================================================================
 UPC - EETAC - CSD - http://digsys.upc.es
 P11:   An example on how to use the XLCD library of functions. 
 included in the XC8 PLIBs (it must be installed from Microchip web
 
 P12:   An old example on how to program the internal Timer0. But better rewrite 
        it to make it compatible with the CSD style.  

 * A 18.5 s timer with two modes of operation: retriggerable (See for example 
 * the classic chip 74LS122/123) and non-retriggerable (see Chip 74LS121)
Features:
 * Programming in FSM style a target chip PIC18F4520. 
 * The use of the microcontroller Timer0 peripheral in 8-bit mode of operation.
 * C language, coding in FSM style, a dedicated processor approach. 
 * PICDEM2+ board, interrupt-driven Trigger signal (a pushbutton to be connected 
 * at RB0/INT0);
 * Retriggerable signal (RT) by polling the switch at RA4 
 * Timer0 peripheral to count real-time up to 18.5 s
 * 
  
Approach: 
- Step 1: Study the notes available on the web about the design of a Timer using 
VHDL and FPGA/CPLD devices.
    - The "Counter Timer" is going to the the Timer0
    - The control Unit is going to be implemented by software programming 
      the microcontroller.
 
- Step 2: Prepare the hardware, so that adapt the PICDEM2+ board to this 
particular application. For instance:
    - Timer_Out may be a LED, for example the one connected at RB3
    - Retriggerable signal (RT), the switch at RA4
    - The Trigger signal is RB0/INT0
    		
- Step 3: Program the init_system() function, configuring ports, interrupts 
and the Timer0. 

---> For instance, in order to generate a timing period of 18.5 s ===> 
FCLK/4 = 1MHz; PRESCALER = 1/16; TMR0L = (256- 250); 
 a software variable TMR0_Post_scaler = 4625  (int);
    
   18.5 s = 1 us * 16 * 250 * 4625          
                  
This is a TMR0 interrupt every 4 ms, which is going to be used to increment 
the TMR0_Post_scaler variable 
              
From the datasheet: 
T0CON: TIMER0 CONTROL REGISTER:
TMR0ON: Timer0 On/Off Control bit ==> 1 = Enables Timer0; 0 = Stops Timer0

T0CS: Timer0 Clock Source Select bit ==> 0 = Internal instruction cycle clock
(CLKO)  [FOSC/4 = 1MHz]

T08BIT: Timer0 8-Bit/16-Bit Control bit ==> 1 = Timer0 is configured as an 8-bit
timer/counter

PSA: Timer0 Prescaler Assignment bit ==> 0 = Timer0 prescaler is assigned. 
Timer0 clock input comes from prescaler output.

T0PS<2:0>: Timer0 Prescaler Select bits ==> 011 = 1:16 Prescaler value


Timer0_Value = 250 ==> TMR0L = (256 - Timer0_Value)  = 6  

And, finally, the extra variable needed to suit the timer period. It will be 
updated by software at every Timer0 interrupt until the "Post_Scaler_Max_Value" 
is reached:

Timer0_Post_Scaler ==> Post_Scaler_Max_Value = 4625 
 * 

NOTE:The C code exemplified here is not optimised in any sense. The CSD style of 
programming aims clarity and facilitating the comprehension of the basic ideas 
on microcontrollers taking advantage of the previous content on combinational
and sequential systems.  
============================================================================= */

#include <xc.h>     // This file also includes the "pic18f4520.h" which contains
                    //all the microcontroller declarations
#include <stdio.h>
#include <stdlib.h>
#include "xlcd.h"
#include "delays.h"

/* 	============================================================================
        Definitions
        ===================================================================== */
// This structure allows to access to a given bit into an 8-bit register
//#define 	PORTBIT(adr, bit) ( (unsigned)(&adr)*8 + (bit) )

/* This is the state enumeration. Check the state diagram posted on the web to
figure out which is the meaning of every state */
#define 	Idle            'A'
#define 	Setup_Timer0    'B'
#define 	Timing          'C'
#define 	Stop_Timer0     'D'

#define 	Post_Scaler_Max_Value 4625 //You may try other values, for instance 
                                       //1000, which give you 4 s
#define 	Timer0_Value 248  // We want to count 250 pulses ... but read the 
//explanation in the Chapter 11 on the way the TMR0 works, when loaded by 
// software .... A couple of cycles are lost every time you reload the timer

/* 	============================================================================
        Function prototypes
    ========================================================================= */
static void interrupt service_routine(void); 
// This is the interrupt service routine

void init_system(void);
char output_logic(void);
char state_logic(void);

void read_inputs(void);  // to interface the hardware pins
void write_outputs(void);

/* 	============================================================================
        Global variables
============================================================================= */

static char Value_RT; 			//Variables to run the FSM
static char Value_Timer_Out;
static char Value_Timer_Out_pre; //previous value
static char current_state = Idle; 	// state variable

static char Trigger_Flag;    	// This flag is going to be set by interrupt 
								//when clicking the trigger push button
static char LCD_Flag;   // This flag is for writing to the LCD 
						//display only when a new message is available.

static char TC_Flag;   // This flag is going to be set once the Timer0,
	// including the "software" postscaler, reaches the Post_Scaler_Max_Value

//The "software" Timer0 post scaler variable to implement a larger counter
static unsigned int Timer0_Post_Scaler; 


/* 	============================================================================
        Main function
        ======================================================================*/
void main(void){
    init_system ();
    while(1) {                      /* loop forever the FSM */
	if (state_logic()) break;
	if (output_logic()) break;
    }
    while (1){
        __nop();// The system has crashed in an illegal state and needs 
                //attention. This loop is never reached in normal operation. 
                //The only way to scape it is clicking the CD
    }   
}

/* 	============================================================================
        Function definitions
        ===================================================================== */

//Future_state_logic routine 
//******************************************************************************
char state_logic(void) {
    char error = 0;
    read_inputs (); // Sample the variables that are going to decide where to go 

    switch (current_state) {
        case Idle:
            if (Trigger_Flag == 1) { // if a button pressed is detected,
                current_state = Setup_Timer0;
                Trigger_Flag = 0;    //Clear variable, 
								//so that it can be interrupted again
                LCD_Flag = 0;
            } else
                current_state = Idle;
            break;

        case Setup_Timer0:
            current_state = Timing;
            break;


        case Timing:
            if (TC_Flag == 1) // End of the timing period
                current_state = Stop_Timer0;

            else if (Trigger_Flag == 1 && Value_RT == 1) { //retrigger the timer
                current_state = Setup_Timer0;
                Trigger_Flag = 0;
            }
            else
                current_state = Timing;
            break;

        case Stop_Timer0:
            current_state = Idle;
            Trigger_Flag = 0; 
                        //Clear variable, so that it can be interrupted again
            break;
        default:
            error = 1;
    }
    return (error);
}

//Output logic routine 
//******************************************************************************
char output_logic(void) {
    char error = 0;
    switch (current_state) {
        case Idle:
            Value_Timer_Out = 0;
            if (LCD_Flag == 1) {	//write only if a new message
                while(BusyXLCD());  WriteCmdXLCD(0x01);     // Line 1
                putrsXLCD("IDLE. Push      ");  
                while(BusyXLCD());  SetDDRamAddr(0x40);     // Line 2
                putrsXLCD("Trigger to start");                              
            }
            LCD_Flag = 0;

            break;

        case Setup_Timer0:
            Value_Timer_Out = 1;

            TMR0IE = 1; // Enable TMR0 interrupts
            TMR0ON = 1; // Start Timer0
            PSA = 0;    // Prescaler in use
            T0PS0 = 1;
            T0PS1 = 1;
            T0PS2 = 0; // T0PS(2..0) = "011" programs a 1:16 Prescaler value

            TMR0L = (256 - Timer0_Value);
            Timer0_Post_Scaler = 0;

            LCD_Flag = 1;   //Write the next message
            break;

        case Timing:
            if (LCD_Flag == 1) {//write the LCD only if there new information
                while(BusyXLCD());  WriteCmdXLCD(0x01);     // Line 1
                putrsXLCD("Timing up to ...");  
                while(BusyXLCD());  SetDDRamAddr(0x40);     // Line 2
                if (Value_RT == 1)
                    putrsXLCD(". 18.5 s    (RT)");                
                else
                    putrsXLCD(". 18.5 s   (nRT)");                        
            }
            LCD_Flag = 0;
            Value_Timer_Out = 1;
            break;

        case Stop_Timer0:
            Value_Timer_Out = 0; // Ends the timing period
            TC_Flag = 0; //Clear the flag
            Timer0_Post_Scaler = 0;
            TMR0IE = 0; // Disable TMR0 interrupts
            TMR0ON = 0; // Stop Timer0
            LCD_Flag = 1; // Because the next state is for writing a message
            break;

        default:
            error = 1;
    }
    write_outputs();// Write the output pins once the output variables are set
    return (error);
}

// Interrupt Service Routine to attend the pushbutton and the Timer0
//******************************************************************************
static void interrupt service_routine(void) { //Interrupt is the key word here
    GIE = 0; //Disable interrupts while attending one of them ...
    if (TMR0IF == 1) {
            // Reload the TMR0 initial value. This a software load, thus some 
            // OSC_CLK pulses will be waisted, meaning time inaccuracy.
            // Is a way to have more time precision? 
        TMR0L = (256 - Timer0_Value); 
        if (Timer0_Post_Scaler < Post_Scaler_Max_Value) {
            Timer0_Post_Scaler++;
        } else {
            TC_Flag = 1; // Set the state diagram variable
        } //(It means that 18.5 s has elapsed (in non retriggerable mode)

        TMR0IF = 0; 
       // Clear the interrupt flag, so that it will be interrupted again in 4 ms
    }
    if (INT0IF == 1) {
        Trigger_Flag = 1; //Set the variable to move the state diagram and clear 
        // the hardware interrupt flag bit (so that the system can be 
        // interrupted again when another active edge is detected)
        INT0IF = 0;
    }

    GIE = 1; 	//Enable interrupts
}

/* =============================================================================
 Scan inputs and process the bits so that we can set our convenient variables. 
 Many of the variables are "char" because we can easily monitor them using
 the watch window for debugging purposes.
  =========================================================================== */
void read_inputs (void){     
// Like in P9, read the port, mask the bit of interest and set the variable 
    Value_RT = (PORTA & 0b00010000)>>4;
   
}

/* =============================================================================
	Driving the output pins. Let's drive the output ports, converting 
	microcontroller variables into electrical signals.    
   ========================================================================== */
void write_outputs(void){
	char Value_data; //Local variable buffer
    
    // Pin RB3 at the PICDEM2+ will be the LED to indicate when the timer is ON
    
    // Indeed, the same way we have implemented an LCD_Flag to write to the 
    // peripheral only when there is a new message to print, there is no need to
    // go rewriting the PORT pins in each loop iteration. Thus, we can consider
    // rewriting the PORT only when the output variables has changed. 
    
    if (Value_Timer_Out != Value_Timer_Out_pre){
	//To write a given output pin:
    // Read and mask to preserve all the bits not to be written
        Value_data = PORTB & 0b11110111;         
    // Shift the bit of interest to the pin position, compose the port bits OR)
    // and write the port
        PORTB = Value_data | (Value_Timer_Out << 3);
        Value_Timer_Out_pre = Value_Timer_Out; 
        // to save the value for the next iteration
    }
}

//*****************************************************************************
//Init operations  
//*****************************************************************************

void init_system(void) {
// Microcontroller general configuration bits. No need to change them.
// PIC18F4520 Configuration Bit Settings. They can be generated automatically 
// in the MPLAB X IDE

// CONFIG1H
#pragma config OSC = XT       // Oscillator Section bits (XT oscillator)
#pragma config FCMEN = OFF     
// Fail-Safe Clock Monitor E bit (Fail-Safe Clock Monitor disabled)
#pragma config IESO = OFF       
// Internal/External Osc Switch-over bit (Oscillator Switch-over mode disabled)

// CONFIG2L
#pragma config PWRT = ON       // Power-up Timer E bit (PWRT enabled)
#pragma config BOREN = SBORDIS 
// Brown-out Reset E bits (Brown-out Reset enabled in hardware 
// only (SBOREN is disabled))
#pragma config BORV = 3    // Brown Out Reset Voltage bits (Minimum setting)

// CONFIG2H
#pragma config WDT = OFF
// Watchdog Timer E bit (WDT disabled (control is on the SWDTEN bit))
#pragma config WDTPS = 32768    
// Watchdog Timer Postscale Sect bits (1:32768)

// CONFIG3H
#pragma config CCP2MX = PORTC  
// CCP2 MUX bit (CCP2 input/output is multiplexed with RC1)
#pragma config PBADEN = OFF    
// PORTB A/D E bit (PORTB<4:0> pins are configured 
// as digital I/O on Reset)
#pragma config LPT1OSC = OFF   
// Low-Power Timer1 Oscillator E bit (Timer1 configured for 
// higher power operation)
#pragma config MCLRE = ON
// MCLR Pin E bit (MCLR pin enabled; RE3 input pin disabled)
 
// CONFIG4L
#pragma config STVREN = ON      
// Stack Full/Underflow Reset E bit (Stack full/underflow will cause Reset)
#pragma config LVP = OFF
// Single-Supply ICSP E bit (Single-Supply ICSP disabled)
#pragma config XINST = OFF
// Extended Instruction Set E bit (Instruction set extension and Indexed 
// Addressing mode disabled (Legacy mode))

// CONFIG5L
#pragma config CP0 = OFF
// Code Protection bit (Block 0 (000800-001FFFh) not code-protected)
#pragma config CP1 = OFF        
// Code Protection bit (Block 1 (002000-003FFFh) not code-protected)
#pragma config CP2 = OFF        
// Code Protection bit (Block 2 (004000-005FFFh) not code-protected)
#pragma config CP3 = OFF        
// Code Protection bit (Block 3 (006000-007FFFh) not code-protected)

// CONFIG5H
#pragma config CPB = OFF        
// Boot Block Code Protection bit 
// (Boot block (000000-0007FFh) not code-protected)
#pragma config CPD = OFF        
// Data EEPROM Code Protection bit (Data EEPROM not code-protected)

// CONFIG6L
#pragma config WRT0 = OFF       
// Write Protection bit (Block 0 (000800-001FFFh) not write-protected)
#pragma config WRT1 = OFF       
// Write Protection bit (Block 1 (002000-003FFFh) not write-protected)
#pragma config WRT2 = OFF       
// Write Protection bit (Block 2 (004000-005FFFh) not write-protected)

#pragma config WRT3 = OFF       
// Write Protection bit (Block 3 (006000-007FFFh) not write-protected)

// CONFIG6H
#pragma config WRTC = OFF       
// Configuration Register Write Protection bit (Configuration registers 
// (300000-3000FFh) not write-protected)
#pragma config WRTB = OFF
// Boot Block Write Protection bit (Boot block (000000-0007FFh) 
// not write-protected)
#pragma config WRTD = OFF
// Data EEPROM Write Protection bit (Data EEPROM not write-protected)

// CONFIG7L
#pragma config EBTR0 = OFF      
// Table Read Protection bit (Block 0 (000800-001FFFh) not protected from 
// table reads executed in other blocks)
#pragma config EBTR1 = OFF      
// Table Read Protection bit (Block 1 (002000-003FFFh) not protected from 
// table reads executed in other blocks)
#pragma config EBTR2 = OFF      
// Table Read Protection bit (Block 2 (004000-005FFFh) not protected from table 
// reads executed in other blocks)
#pragma config EBTR3 = OFF      
// Table Read Protection bit (Block 3 (006000-007FFFh) not protected from table 
// reads executed in other blocks)

// CONFIG7H
#pragma config EBTRB = OFF      
// Boot Block Table Read Protection bit (Boot block (000000-0007FFh) 
// not protected from table reads executed in other blocks)

    /* -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+*/
	
// Initialise starts here:
    // By default, let's declare all the unused pins as digital outputs 
   
// The ports:
    PORTA = 0x00;   // Reset all Flip-Flops at PORTA
    LATA = 0x00;    // Reset all Flip-Flops at PORTA
    ADCON1 = 0x0F;  // Configures all the PORTA and PORTB(3..0) lines as digital
    TRISA = 0b00010000; // RA4 is the RT input
    
    PORTB = 0x00; // Reset all Flip-Flops at PORTB
    LATB = 0x00; // Reset all Flip-Flops at PORTB
    TRISB = 0b00000001; // PORTB RB3 is the Timer_Out output
                        // PORTB RB0 is the INT0 trigger input

// PORTD is used by the LCD 
// It's initialised in the init_XLCD(); 
    
// The TMR0
    TMR0ON = 0;     // Timer0 stopped
    T0CS = 0;       // Internal clock (CLKO)  [FOSC/4 = 1MHz]
    T08BIT = 1;     // Timer0 is configured as an 8-bit timer/counter
    PSA = 0;        // Prescaler in use
    T0PS0 = 1;
    T0PS1 = 1;
    T0PS2 = 0;      // T0PS(2..0) = "011" programs a 1:16 Prescale value
// and the TMR0 interrupt:    
    TMR0IE = 0;     // Disable TMR0 interrupts at Idle state
    
// The external interrupt INT0
    INTEDG0 = 0;    // INT0 edge selection: '0' --> on falling edge 
                    //(interrupt after clicking the pushbutton)
    INT0IE = 1;     // Enable interrupts from the INT0 pin

// Let's allow interrupts     
    GIE = 1;       

//Initial situation at "Idle" state:
    Value_Timer_Out = 0;
    Trigger_Flag = 0;
    TC_Flag = 0;

// LCD initialisation and some messages to see how it works
    OpenXLCD(FOUR_BIT & LINES_5X7); //configure LCD in 4-bit Data Interface mode
                                    //and 5x7 characters, multiple line display
    while(BusyXLCD());  WriteCmdXLCD(0x06);   // move display left          
                        
    while(BusyXLCD());  WriteCmdXLCD(0x0C);  //turn display on without cursor
                 
    putrsXLCD("CSD -- P11 - P12");  //Display the ASCII String
    SetDDRamAddr(0x40);             //shift cursor to beginning of second line
    putrsXLCD("  LCD Display   ");  //Display "LCD display"

    Delay10KTCYx(150);         // Stop the LCD after 1.5 seconds
   
    // Flip the display messages  
    while(BusyXLCD());  WriteCmdXLCD(0x01);  // Go line 1 clearing the display
    putrsXLCD("  LCD Display   ");  
    
    while(BusyXLCD());  SetDDRamAddr(0x40); // Go second line 
    putrsXLCD("CSD -- P11 - P12");  
                
    Delay10KTCYx(250);  
    LCD_Flag = 1; //Ready for printing the messages in the next Idle state
}